#include "6821.h"
#include "w_osi.h" //TRACE

extern BOOL is_peek_access;

// 6821/6520 PIA
//
//
//
//            +-----+  +-----+
//       Vss  |  1  +--+  40 | CA1
//       AA0  |  2        39 | CA2
//       PA1  |  3        38 | !IRQA
//       PA2  |  4        37 | !IRQB
//       PA3  |  5        36 | RS0
//       PA4  |  6        35 | RS1
//       PA5  |  7        34 | !RES
//       PA6  |  8        33 | D0
//       PA7  |  9        32 | D1
//       PB0  | 10        31 | D2
//       PB1  | 11        30 | D3
//       pB2  | 12        29 | D4
//       PB3  | 13        28 | D5
//       PB4  | 14        27 | D6
//       PB5  | 15        26 | D7
//       PB6  | 16        25 | ENABLE
//       PB7  | 17        24 | CS2
//       CB1  | 18        23 | !CS3
//       CB2  | 19        22 | CS1
//       Vcc  | 20        21 | R/W
//            +--------------+
//

/*
The PIA has two 8-bit I/O ports, A and B, which are mostly identical in
function.  All 8 bits can be set to input (1 in the DDR) or output (0 in
the DDR) independently. Bit 2 in CRx determines whether the DDRx or Px are
accessed.  Each port has 2 control lines: CA1, CA2, CB1, CB2. The Cx1 are
input only, the Cx2 can be input or output.

PIAs have two interrupt lines: IRQA and IRQB. They may go low on a change
on the inputs of the Cxy lines. These interrupts may be enabled on either
an 1->0 or an 0->1 transition. The selected transition is called the
"active" transition. The flags which register that an active transition has
occurred are reset by reading the appropriate PORT register.

Register map:

0x0  PORT A or DDR A: Data Direction Register A
0x1  CRA: Control Register A
0x2  PORT B or DDR B: Data Direction Register B
0x3  CRB: Control Register B

Control registers:

CRA:

bit   meaning
---   -------
 7    CA1 active transition flag. 1= 0->1, 0= 1->0
 6    CA2 active transition flag. 1= 0->1, 0= 1->0
 5    CA2 direction           1 = out        | 0 = in
                    ------------+------------+---------------------
 4    CA2 control   Handshake=0 | Manual=1   | Active: High=1 Low=0
 3    CA2 control   On Read=0   | CA2 High=1 | IRQ on=1, IRQ off=0
                    Pulse  =1   | CA2 Low=0  |
 2    Port A control: DDRA = 0, PORTA = 1
 1    CA1 control: Active High = 1, Low = 0	 CA1 bit7 set on H->L(0) or L->H(1)
 0    CA1 control: IRQ on=1, off = 0

CRB works identical for CB1 and CB2, except for the differences in
handshaking.

The Cx2 handshake is not identical between ports. For port A, the handshake
is on reading the PORT A register, for CB2 the handshake is sent on writing
the PORT B register.

On the listening side:

BIT 3 LOW with CA2: CA2 is now controlled by two events:
    (i) CA1 active transition sets it high ("Data Valid")
    (ii) a READ operation sets it low ("Data Accepted").

On the talking side:

BIT 3 LOW with CB2: CB2 is now controlled by two events:
    (i) CB1 active transition sets it high ("Ready For Data")
    (ii) a WRITE operation sets it low ("Data Valid").

Bit 2 HIGH: Causes pulse output, CA2 or CB2 going low for one cycle only
after a read or write operation. This pulse may be too short for some uses.

For operation with a sending and a receiving PIA one would connect

talker      listener
------      --------
Port B   -> Port A
CB2      -> CA1 with active = 1->0 (data valid)
CB1     <-  CA2 with active = 1->0 (data accepted and ready for more)

*/

PIA6821::PIA6821()
{
	Reset();
}

PIA6821::~PIA6821()
{
}

void PIA6821::Reset()
{
   /* clear _all_ internal registers */

   ctrl_a = 0;    /* PIA Port A Control */
   ctrl_b = 0;    /* PIA Port B Control */
   ddr_a = 0;     /* PIA Port A DDR */
   ddr_b = 0;     /* PIA Port B DDR */
   port_a = 255;  /* PIA Port A input */
   port_b = 255;  /* PIA Port B input */

   ca_state = 1;
   cb_state = 1;
   m_nIRQ = 0;
}

void PIA6821::UpdateIRQ() 
{
    if (   ((ctrl_a & 0x81) == 0x81) ||
           ((ctrl_a & 0x68) == 0x48) ||
           ((ctrl_b & 0x81) == 0x81) ||
           ((ctrl_b & 0x68) == 0x48) 
		)
	{
        m_nIRQ = 1;
		//my_set_int(1);
    } 
	else 
	{
        m_nIRQ = 0;
		//my_set_int(0);
    }
}


void PIA6821::Write( unsigned int address, unsigned char writevalue)
{

	switch (address & 3)
	{
		case P_PORT_A:	//PORTA v DDRA
			if (ctrl_a & 4) 
			{
				//writevalue = (BYTE)(writevalue | ~ddr_a);
				WritePortA(writevalue);
				port_a = writevalue;
			} 
			else 
			{
				ddr_a = writevalue;
			}
			break;

		case P_CTRL_A: //CRA
			if ((writevalue & 0x38) == 0x30 ) 
			{   
				/* set output low */
				SetCA2(0);
				ca_state = 0;
			} 
			else if ((writevalue & 0x38) == 0x38) 
			{   
				/* set output high */
				SetCA2(1);
				ca_state = 1;
			} 
			else if ((ctrl_a & 0x30) == 0x30) 
			{
				/* change to toggle/pulse */
				SetCA2(1);
				ca_state = 1;
			}

			ctrl_a = (BYTE)((ctrl_a & 0xc0) | (writevalue & 0x3f));

			if (ctrl_a & 0x20) 
				ctrl_a &= 0xbf;

			UpdateIRQ();
			break;

		case P_PORT_B:	//PORTB v DDRB
			if (ctrl_b & 4) 
			{
				//writevalue = (BYTE)(writevalue | ~ddr_b);
				WritePortB(writevalue);
				port_b = writevalue;
			} 
			else 
			{
				ddr_b = writevalue;
			}
			
			if (IS_CB2_HANDSHAKE()) 
			{
				SetCB2(0);
				cb_state = 0;
				if (IS_CB2_PULSE_MODE()) 
				{
					SetCB2(1);
					cb_state = 1;
				}
			}
			break;

		case P_CTRL_B: //CRB
			if ((writevalue & 0x38) == 0x30 ) 
			{   
				/* set output low */
				SetCB2(0);
				cb_state = 0;
			} 
			else if ((writevalue & 0x38) == 0x38) 
			{   
				/* set output high */
				SetCB2(1);
				cb_state = 1;
			} 
			else if ((ctrl_b & 0x30) == 0x30) 
			{
				/* change to toggle/pulse */				
				SetCB2(1);
				cb_state = 1;
			}

			ctrl_b = (BYTE)((ctrl_b & 0xc0) | (writevalue & 0x3f));

			if ( ctrl_b & 0x20 ) 
				ctrl_b &= 0xbf;

			UpdateIRQ();

			break;
	}

}
void PIA6821::Read( unsigned int address, unsigned char *readvalue)
{

	switch (address & 3)
	{
		case P_PORT_A:	//PORTA v DDRA
			if (ctrl_a & 4) 
			{

				if(!is_peek_access) 
				{
					ctrl_a &= 0x3f;           /* Clear CA1,CA2 IRQ */
					UpdateIRQ();
				}
				/* WARNING: this pin reads the voltage of the output pins, not
				   the ORA value as the other port. Value read might be different
				   from what is expected due to excessive electrical load on the
				   pin. */
				
				m_bPortAValue = ReadPortA();
				*readvalue = m_bPortAValue;
				return;
			}
			
			m_bPortAValue = ddr_a;
			*readvalue = m_bPortAValue;
			return;
			
			break;

		case P_CTRL_A: //CRA
			*readvalue = ctrl_a;
			return;
			break;
		         

		case P_PORT_B: //PORTB v DDRB

			if (ctrl_b & 4) 
			{

				if(!is_peek_access) 
				{
					ctrl_b &= 0x3f;           /* Clear CB1,CB2 IRQ */
					UpdateIRQ();
				}

				/* WARNING: this pin reads the ORA for output pins, not
				   the voltage on the pins as the other port. */
				m_bPortBValue = ReadPortB();
				m_bPortBValue = (BYTE)((m_bPortBValue & ~ddr_b) | (port_b & ddr_b));
				*readvalue = m_bPortBValue;
				return;
			}
			m_bPortBValue = ddr_b;
			*readvalue = m_bPortBValue;
			return;
			break;

		case P_CTRL_B: //CRB
			*readvalue = ctrl_b;
			if (ctrl_b & 0x80)
			{
				ctrl_b &= 0x7f;
			}
			return;
			break;
		        
	}

}

void PIA6821::WritePortA(BYTE )
{
	//base class do nothing
}
void PIA6821::WritePortB(BYTE )
{
	//base class do nothing
}

BYTE PIA6821::ReadPortA()
{
	return 0;
}

BYTE PIA6821::ReadPortB()
{
	return 0;
}

void PIA6821::SetCA1(BYTE )
{
	//base class do nothing
	TRACE("SetCA1 = 0x%0x\n");
}
void PIA6821::SetCA2(BYTE )
{
	//base class do nothing
	TRACE("SetCA2 = 0x%0x\n");

}
void PIA6821::SetCB1(BYTE bSet)
{
	//base class do nothing
	//TRACE("SetCB1 = 0x%0x\n", bSet);
	
	ctrl_b &= 0x7f;
	if (bSet)
		ctrl_b |= 0x80;

}
void PIA6821::SetCB2(BYTE )
{
	//base class do nothing
	TRACE("SetCB2 = 0x%0x\n");

}

