 // 6850.cpp: implementation of the Acia class.
//
//////////////////////////////////////////////////////////////////////

#include "6850.h"
#include "w_osi.h"		//report system error
#include "fileio.h"		//sztFileHome
#include "resource.h"
#include <TCHAR.H>
#include <windowsx.h>

const TCHAR szSection[]		= _T("SERIAL");
const TCHAR szSpeedEntry[]	= _T("Speed");
const TCHAR szLOADEntry[]	= _T("Autoload");
const TCHAR szSAVEEntry[]	= _T("Autosave");
const TCHAR szPathEntry[]   = _T("LastPath");

//
// 6850
//
//                      6850
//                +------\/------+
//             1 -|Vss       !CTS|- 24
//             2 -|RxD       !DCD|- 23
//             3 -|RxClk       D0|- 22
//             4 -|TxClk       D1|- 21
//             5 -|!RTS        D2|- 20
//             6 -|TxD         D3|- 19
//             7 -|!IRQ        D4|- 18
//             8 -|CS0         D5|- 17
//             9 -|!CS2        D6|- 16
//            10 -|CS1         D7|- 15
//            11 -|RS           E|- 14
//            12 -|Vcc        R/W|- 13
//                +--------------+ 
//
// address 0 read	 = data bits 0-7
// address 0 write   = data bits 0-7
//
// address 1 write
// DB0 = divide sel 0     00 = /1	 10 = /64
// DB1 = divide sel 1	  01 = /16	 11 = master reset
// DB2 = word sel 0		 000 = 7e2	011 = 7o1		110 = 8e1
// DB3 = word sel 1		 001 = 7o2	100 = 8n2		111 = 8o1
// DB4 = word sel 2		 010 = 7e1	101 = 8n1
// DB5 = transmit ctrl 0 00 = -RTS low xmit irq disable
// DB6 = transmit ctrl 1 01 = -RTS low xmit irq enable
//                       10 = -RTS high xmit irq disable
//                       11 = -RTS low, xmit a break lvl, irq disable
// DB7 = receive irq enable
//
// address 1 read
// DB0 = RX buff full
// DB1 = TX buff empty
// DB2 = -DCD			(1 when DCD not present)
// DB3 = -CTS			(1 when transmit is inhibited)
// DB4 = Framing error	(1 when error, lack of start/stop bits or break condition)
// DB5 = RX overrun
// DB6 = Parity Error   (1 when # of 1 bits != selected parity)
// DB7 = IRQ            (1 when IRQ present)  cleared by read of rx or write to tx
//
// rxIRQ is set when Receive Data Register is full, overrun, or there is a low to high 
// transition on the DCD signal line.
//
// txIRQ is set when the Transmit Data Register is empty,


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Acia::Acia( int nUnit /*=0*/) : m_nUnit(nUnit)
{
	hInFile = INVALID_HANDLE_VALUE;
	hOutFile = INVALID_HANDLE_VALUE; 
	m_eSpeed = bNoLimit; //no limit

	m_dwRTickCount = 0;
	m_dwRNextTick = 0;
	m_dwTTickCount = 0;
	m_dwTNextTick = 0;

	
	m_cOut = 0;   
	m_txIRQ = FALSE;      
	m_rxIRQ = FALSE;      
	m_txIRQenable = FALSE;
	m_rxIRQenable = FALSE;

	m_bAutoSave = FALSE;
	m_bAutoLoad = FALSE;
	 
}

Acia::~Acia()
{
	
	if (hInFile && hInFile != INVALID_HANDLE_VALUE)
		CloseHandle(hInFile);

	if (hOutFile && hOutFile != INVALID_HANDLE_VALUE)
		CloseHandle(hOutFile);

}

void Acia::Reset()
{
	m_eDivideSel	= DivideBy1;
	m_eWordSel		= SevenE2;
	m_eXmitCtrl		= RTSNOIRQ;
	m_rxIRQenable	= FALSE;
	m_txIRQenable	= FALSE;
}

void Acia::Write( unsigned int address, unsigned char writevalue)
{
	if ( (address & 1) )
	{
		data_write( writevalue );
	}
	else
	{
		control_write( writevalue);
	}
}

void Acia::Read( unsigned int address, unsigned char *readvalue)
{
	if ( address & 1)
	{
		*readvalue = data_read();
	}
	else
	{
		*readvalue = control_read();

	}
}

// To be used to update ACIA internal state
void Acia::clock_tick()
{

}


// control read
// DB0 = RX buff full	1=full
// DB1 = TX buff empty	1=empty
// DB2 = -DCD
// DB3 = -CTS					 (
// DB4 = Framing error	1=FE	 (never set)
// DB5 = RX overrun		1=overrun
// DB6 = Parity Error
// DB7 = IRQ
// eStatus { rxFull = 1,
//			 txFull = 2,
//			 inDCD = 4,
//			 inCTS = 8,
//			 FrameError = 16,
//			 rxOverrun = 32,
//			 rxParityError = 64,
//			 rxIRQ = 128; };
	

BYTE Acia::control_read()
{
		
	m_cStatus &= ~acia_RXFull; //rxEmpty

	if (m_dwRTickCount)	//receive pacing on
	{
		m_dwRTickCount = GetTickCount();
		if (m_dwRTickCount < m_dwRNextTick )
		{
			//we haven't waited long enough for next character
			//m_cStatus &= ~acia_RXFull;	
		}
		else
		{
			//there is a new character to receive
			if (hInFile != INVALID_HANDLE_VALUE || m_bAutoLoad)
				m_cStatus |= acia_RXFull;	
		}
	}
	else
	{
		//we're always ready to read next character
		//if there are any
		if (hInFile != INVALID_HANDLE_VALUE || m_bAutoLoad)
			m_cStatus |= acia_RXFull;
	}



	if (m_dwTTickCount)	//transmit pacing on
	{
		
		m_dwTTickCount = GetTickCount();
		if (m_dwTTickCount < m_dwTNextTick )
		{
			//but it hasn't been shifted out yet....
		}
		else
		{
			//time to receive new character
			m_cStatus |= acia_TXEmpty;
		}
	}
	else
	{
		if ( (m_cStatus & acia_TXEmpty) == 0)
		{
			//we're always ready to receive next character
			m_cStatus |= acia_TXEmpty;
			
			if (m_txIRQenable)
				m_txIRQ = TRUE; //TODO: handle IRQ set
		}
	}

	if (m_txIRQ || m_rxIRQ)
		m_cStatus |= acia_IRQ; 

	return m_cStatus;
}

// control write
// DB0 = divide sel 0    00= /1	   10= /64
// DB1 = divide sel 1	   01= /16	 11= master reset
// DB2 = word sel 0		   000 = 7e2	011 = 7o1		110 = 8e1
// DB3 = word sel 1		   001 = 7o2	100 = 8n2		111 = 8o1
// DB4 = word sel 2		   010 = 7e1	101 = 8n1
// DB5 = transmit ctrl 0 00 = -RTS low xmit irq disable
// DB6 = transmit ctrl 1 01 = -RTS low xmit irq enable
//                       10 = -RTS high xmit irq disable
//                       11 = -RTS low, xmit a break lvl, irq disable
// DB7 = receive irq enable
//
void Acia::control_write(BYTE data)
{
	int val;
	switch (data & 0x03)
	{
		case 0:
			// divide by 1
			m_eDivideSel = DivideBy1;
			break; 
		case 1:
			// divide by 16
			m_eDivideSel = DivideBy16;
			break;
		case 2:
			//divide by 64
			m_eDivideSel = DivideBy64;
			break;
		case 3:
			Reset();
			return;
			break;
	}

	//set word size, parity, stopbits
	val = (data & 0x1C) >> 2;
	m_eWordSel = (eWordSel) val;

	//set RTS, xmit IRQ, Break
	val = (data & 0x60) >> 5;
	m_eXmitCtrl = (eXmitCtrl) val;

	m_rxIRQenable = (data & '\x80') != 0;
	m_txIRQenable = ((data & '\x60') == '\x20');
		
}

// read last character received.
// reading clears irq read status register
BYTE Acia::data_read()
{
	
	DWORD dwRead  = 0;
	BYTE mask = 0xFF;

	//read operation clear Flags
	m_cStatus &= ~(acia_RXFull | acia_IRQ | acia_PE | acia_FE  | acia_RXO);
	
	if (m_bAutoLoad && hInFile == INVALID_HANDLE_VALUE)
	{
		SetFile(TRUE);
		
		//user didn't really want to save output
		if (hInFile == INVALID_HANDLE_VALUE)
		{
			//Need to determine how to prevent auto open from occuring
			//at this point
			
			//For now, just set no characters time to 15 seconds
			m_dwRTickCount = GetTickCount();
			m_dwRNextTick = m_dwRTickCount + 10000;

		}
	}	
	
	
	if ( m_eWordSel  < EightN2)
		mask = 0x7F;

	
	if (hInFile != INVALID_HANDLE_VALUE)
	{
		
		//pacing is on, return last byte again
		if (m_dwRTickCount && m_dwRNextTick > GetTickCount())
			return (BYTE) (m_cOut & mask);
		
		if (!ReadFile( hInFile, &m_cOut, 1, &dwRead, NULL))
		{
			DWORD dwError = GetLastError();
			ReportSystemError(dwError);

			if (dwError == ERROR_HANDLE_EOF)
			{
				CloseHandle(hInFile);
				hInFile = INVALID_HANDLE_VALUE;
				m_dwRTickCount = 0;
				m_cOut = 0;

			}

		}
		else
		{
			

			
			
			if (dwRead == 0)
			{
				//no more characters to read.  To bad we didn't know this before
				//flagging character ready. return a null character	from now on
				m_cOut = 0;

				CloseHandle(hInFile);
				hInFile = INVALID_HANDLE_VALUE;

				//Need to determine how to prevent auto open from occuring
				//at this point
				
				//For now, just set no characters time to 15 seconds
				m_dwRTickCount = GetTickCount();
				m_dwRNextTick = m_dwRTickCount + 10000;
			}
			else if (m_eSpeed)
			{
				
				
				m_dwRTickCount = GetTickCount();
				//this calculation is wrong since we always wait one
				//byte time after transmitting. Use it until we use CPU cycles
				//to determine rxTime
				m_dwRNextTick = m_dwRTickCount + m_eSpeed;

			}
			else
			{
				m_dwRTickCount = 0;
				m_dwRNextTick = 0;
			}
			


			return (BYTE)(m_cOut & mask);
		}
	}
	else
	{
		//no file opened

	}
	
	return 0;
}

// 6850 has 1 byte buffer essentially.
// after output byte has been stuffed into output shift register,
// the TXEmpty flag is set, allowing 1 byte to be written
// The flag remains set until that byte is moved into the shift register
// (*Notes for future implementation)
//
void Acia::data_write(BYTE data)
{
	//leading bit out is bit 0

	if ( m_eWordSel  < EightN2)
		data &= 0x7F; //7bit mode

	if (m_bAutoSave && hOutFile == INVALID_HANDLE_VALUE)
	{
		SetFile(FALSE);
		
		//user didn't really want to save output
		if (hOutFile == INVALID_HANDLE_VALUE)
			m_bAutoSave = FALSE;
	}
	
	if (hOutFile != INVALID_HANDLE_VALUE)
	{
		DWORD dwWritten;
		if (!WriteFile(hOutFile, &data, 1, &dwWritten, NULL) || dwWritten != 1)
		{
			dwWritten = GetLastError();
			ReportSystemError(dwWritten);
		}

	}

	//set txBuf full status;
	m_cStatus &= ~acia_TXEmpty;


	if (m_eSpeed)
	{

		m_dwTTickCount = GetTickCount();
		m_dwTNextTick = m_dwTTickCount + m_eSpeed;

	}

}

#define TITLELEN 80
#define MAXFILTER 255
void Acia::SetFile(BOOL bInput /*= TRUE*/)
{
	OPENFILENAME ofn;
	TCHAR file[_MAX_PATH];
	TCHAR szTitle[TITLELEN];
	TCHAR szFilter[MAXFILTER];
	TCHAR Path[_MAX_PATH];
	TCHAR szUnitSection[ sizeof(szSection) + 12 ];
	BOOL bOK = FALSE;

	//create unit entry if multiple ports
	_tcscpy( szUnitSection, szSection);
	if (m_nUnit)
	{
		_stprintf(szUnitSection+ _tcslen(szSection), _T("%d"), m_nUnit); 
	}


	if (GetPrivateProfileString(szUnitSection, szPathEntry, _T(""), 
								Path, _MAX_PATH, m_szIniFile) < 2)
	{
		_tcscpy(Path, sztFileHome);
	}
	else
	{
		int nLen = _tcslen(Path);
		if (Path[nLen-1] != _T('\\'))
		{
			TCHAR *p = _tcsrchr(Path, _T('\\'));
			if (p)
				*(p+1) = _T('\0');
		}

	}
	

	ZeroMemory(&ofn, sizeof(OPENFILENAME));

	if (LoadString(ghInstance, IDS_SERIAL_FILTER, szFilter, MAXFILTER))
	{
		for (int i=0; i < MAXFILTER; i++)
		{
			//filter needs null separating paramters, replace | in resource with Null
			if (szFilter[i] == _T('|'))
				szFilter[i] = _T('\0');
		}
	} 


	*file = _T('\0');
	*szTitle = _T('\0');

	if (bInput)
	{
		LoadString(ghInstance, IDS_SERIAL_IN, szTitle, TITLELEN);
	}
	else
	{
		LoadString(ghInstance, IDS_SERIAL_OUT, szTitle, TITLELEN);
	}	

	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = ghWnd_Main;
	ofn.hInstance = ghInstance;
	ofn.Flags = OFN_EXPLORER | (bInput ? OFN_FILEMUSTEXIST : (OFN_OVERWRITEPROMPT | OFN_HIDEREADONLY) );
	ofn.lpstrInitialDir	= Path;
	ofn.lpstrFile = file;
	ofn.nMaxFile = _MAX_PATH;
	ofn.lpstrTitle = szTitle;
	ofn.lpstrFilter = szFilter;

	if (bInput)
	{
		bOK = GetOpenFileName( &ofn );
		if (bOK)
		{
		
			
			if (hInFile && hInFile != INVALID_HANDLE_VALUE)
			{
				//Close old file first
				CloseHandle(hInFile);
				hInFile = INVALID_HANDLE_VALUE;
			}
			
			hInFile = CreateFile(file, GENERIC_READ, FILE_SHARE_READ,
					NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, 
					NULL );

			if (hInFile == INVALID_HANDLE_VALUE)
			{
				ReportSystemError(GetLastError());
				return;
			}
			else
			{
				TCHAR drive[_MAX_DRIVE];
				_tsplitpath( file, drive, Path,	NULL, NULL);
				_tmakepath(file, drive, Path , NULL, NULL);
				m_dwRTickCount = 0;
			}

		}
	}
	else
	{
		bOK = GetSaveFileName( &ofn );
		
		if (bOK)
		{
			if ( hOutFile && hOutFile != INVALID_HANDLE_VALUE)
			{
				//close old file first
				CloseHandle(hOutFile);
				hOutFile = INVALID_HANDLE_VALUE;
			}
			
			
			
			hOutFile = CreateFile(file, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ,
					NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, 
					NULL );

			if (hOutFile == INVALID_HANDLE_VALUE)
			{
				ReportSystemError(GetLastError());
				return;
			}
			else
			{
				TCHAR drive[_MAX_DRIVE];
				_tsplitpath( file, drive, Path,	NULL, NULL);
				_tmakepath(file, drive, Path , NULL, NULL);
				m_dwTTickCount = 0;
			}


		}
		 
	}
	if (bOK && m_szIniFile && *m_szIniFile)
	{
		//save open path
		WritePrivateProfileString(szUnitSection, szPathEntry, file, m_szIniFile);

	}

}


void Acia::SetBaud(eBaud Speed)
{
	
	m_eSpeed = Speed;
	m_dwRTickCount = 0;
	m_dwTTickCount  = 0;


}
BOOL Acia::IsFileOpen(BOOL bInput/* = TRUE*/)
{
	if (bInput)
	{
		if (hInFile != INVALID_HANDLE_VALUE)
			return TRUE;
		return FALSE;
	}

	if (hOutFile != INVALID_HANDLE_VALUE)
		return TRUE;

	return FALSE;
}
//close file if open
void Acia::CloseFile(BOOL bInput /*=TRUE */)
{
			
	if (bInput)
	{
		if (hInFile && hInFile != INVALID_HANDLE_VALUE)
		{
			//Close old file first
			CloseHandle(hInFile);
			hInFile = INVALID_HANDLE_VALUE;
		}
	}
	else
	{
		if ( hOutFile && hOutFile != INVALID_HANDLE_VALUE)
		{
			//close old file first
			CloseHandle(hOutFile);
			hOutFile = INVALID_HANDLE_VALUE;
		}

	}

}

//Set .ini file for other operations
void Acia::SetINIFile(LPCTSTR szFile)
{
	*m_szIniFile = _T('\0');
	if (_tcslen(szFile) < _MAX_PATH)
		_tcscpy( (LPTSTR) m_szIniFile, szFile);

}
//Read .ini file settings, use optional file spec
BOOL Acia::LoadConfig( LPCTSTR szFile /*= NULL*/ )
{
	int nVal;
	TCHAR szUnitSection[ sizeof(szSection) + 12 ];

	if (szFile && *szFile)
	{
		SetINIFile(szFile);
	}

	//see if INI file exists
	if ( !*m_szIniFile || GetFileAttributes(  m_szIniFile) == (DWORD) -1 )
		return FALSE;

	//create unit entry if multiple ports
	_tcscpy( szUnitSection, szSection);
	if (m_nUnit)
	{
		_stprintf(szUnitSection+ _tcslen(szSection), _T("%d"), m_nUnit); 
	}


	nVal = GetPrivateProfileInt(szUnitSection, szSpeedEntry, 0, m_szIniFile);

	if (nVal == 0)
		SetBaud(bNoLimit);
	else if (nVal <= 110)
		SetBaud(b110);
	else if (nVal <= 300)
		SetBaud( b300 );
	else if (nVal <= 600)
		SetBaud( b600 );
	else if (nVal <=1200)
		SetBaud( b1200 );
	else if (nVal <= 2400 )
		SetBaud( b2400 );
	else if (nVal <=4800 )
		SetBaud( b4800 );
	else if (nVal <=9600 )
		SetBaud( b9600 );
	else
		SetBaud (bNoLimit );

	nVal = GetPrivateProfileInt(szUnitSection, szLOADEntry, 0, (LPCTSTR)m_szIniFile);
	m_bAutoLoad = nVal > 0 ? TRUE : FALSE;
	
	nVal = GetPrivateProfileInt(szUnitSection, szSAVEEntry, 0, (LPCTSTR)m_szIniFile);
	m_bAutoSave = nVal > 0 ? TRUE : FALSE;  
	  
	return TRUE;		

}

//Save .ini file settings, use option file spec
BOOL Acia::SaveConfig( LPCTSTR szFile /*= NULL*/ )
{
	
	BOOL bOK = TRUE;
	TCHAR *p = NULL;
	TCHAR szUnitSection[ sizeof(szSection) + 12 ];

	_tcscpy( szUnitSection, szSection);
	 
	if (m_nUnit)
	{
		_stprintf(szUnitSection+ _tcslen(szSection), _T("%d"), m_nUnit); 
	}
	
	if (szFile && *szFile)
	{
		SetINIFile(szFile);
	}

	//see if INI file exists
	if ( !*m_szIniFile || GetFileAttributes( m_szIniFile) == (DWORD) -1 )
		return FALSE;


	switch (GetBaud())
	{
		case b110:
			p = _T("110");
			break;
		case b300:
			p = _T("300");
			break;
		case b600:
			p = _T("600");
			break;
		case b1200:
			p = _T("1200");
			break;
		case b2400:
			p = _T("2400");
			break;
		case b4800:
			p = _T("4800");
			break;
		case b9600:
			p = _T("9600");
			break;
		case bNoLimit:
		default:
			p = _T("0");
			break;
	}
	
	bOK = WritePrivateProfileString(szUnitSection, szSpeedEntry, p, m_szIniFile); 

	if (bOK)
		bOK = WritePrivateProfileString(szUnitSection, szLOADEntry, 
			m_bAutoLoad ? _T("1") : _T("0"), m_szIniFile); 
	if (bOK)
		bOK = WritePrivateProfileString(szUnitSection, szSAVEEntry, 
			m_bAutoSave ? _T("1") : _T("0"), m_szIniFile); 

	return bOK;


}

//*******************************************************************
// SerialDlg - static member, handle Dialog messages
//
// paramaters:
//             hdlg          - The window handle for this message
//             msg	        - The message number
//             wparam        - The WPARAM parameter for this message
//             lparam        - The LPARAM parameter for this message
//
//*******************************************************************
BOOL CALLBACK  Acia::SerialDlg(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{

	static Acia* pACIA = NULL;
	int nIndex;

	switch(msg)
	{
		case WM_INITDIALOG:
			{
				if (!lParam)
				{
					EndDialog(hDlg, IDABORT);
					return FALSE;
				}
				//pointer to ACIA passed by lParam
				pACIA = (Acia*)	lParam;
				CenterWindow( hDlg, (HWND) ghWnd_Main); 

				switch(pACIA->GetBaud())
				{
					case b110:
					case b300:
						nIndex = 0;
						break;
					case b600:
						nIndex = 1;
						break;
					case b1200:
						nIndex = 2;
						break;
					case b2400:
						nIndex = 3;
						break;
					case b4800:
						nIndex = 4;
						break;
					case b9600:
						nIndex = 5;
						break;
					case bNoLimit:
					default:
						nIndex = 6;
						break;
				}
				CheckRadioButton (hDlg, IDC_RB_300, IDC_RB_NOLIMIT, IDC_RB_300 + nIndex);
				CheckDlgButton (hDlg, IDC_CB_AUTOSAVE, pACIA->GetAutoSave() ? 1 : 0);
				CheckDlgButton( hDlg, IDC_CB_AUTOLOAD, pACIA->GetAutoLoad() ? 1 : 0);
				return FALSE; //haven't set focus to control
			}
			break;

		case WM_COMMAND:
			switch (GET_WM_COMMAND_ID(wParam, lParam))
			{
				case IDOK:
					{
						eBaud NewSpeed =
								   IsDlgButtonChecked (hDlg, IDC_RB_300) ? b300 :
								   IsDlgButtonChecked (hDlg, IDC_RB_600) ? b600 : 
								   IsDlgButtonChecked (hDlg, IDC_RB_1200) ? b1200 : 
								   IsDlgButtonChecked (hDlg, IDC_RB_2400) ? b2400 :
								   IsDlgButtonChecked (hDlg, IDC_RB_4800) ? b4800 :
								   IsDlgButtonChecked (hDlg, IDC_RB_9600) ? b9600 : bNoLimit;

						if (NewSpeed != pACIA->GetBaud())
							pACIA->SetBaud(NewSpeed);
						
						pACIA->SetAutoSave( IsDlgButtonChecked (hDlg, IDC_CB_AUTOSAVE) ? TRUE : FALSE );
						pACIA->SetAutoLoad( IsDlgButtonChecked (hDlg, IDC_CB_AUTOLOAD) ? TRUE : FALSE );					
						pACIA->SaveConfig();
						EndDialog(hDlg, IDOK);
						
					}
					break;
				case IDCANCEL:
					EndDialog(hDlg, IDCANCEL);
					break;
				default:
					return FALSE;
			}
		default:
			return(FALSE);
	}
	return(FALSE);
}
