// 6850.h: interface for the Acia class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_ACIA_H__EE7A53A0_A454_11D5_B5C9_00E01890870B__INCLUDED_)
#define AFX_ACIA_H__EE7A53A0_A454_11D5_B5C9_00E01890870B__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "6502.h"
#include <windows.h>
#include <TCHAR.H>
//
// 6850
//
//           6850
//     +------\/------+
//  1 -|Vss       /CTS|- 24
//  2 -|RxD       /DCD|- 23
//  3 -|RxClk       D0|- 22
//  4 -|TxClk       D1|- 21
//  5 -|/RTS        D2|- 20
//  6 -|TxD         D3|- 19
//  7 -|/IRQ        D4|- 18
//  8 -|CS0         D5|- 17
//  9 -|/CS2        D6|- 16
// 10 -|CS1         D7|- 15
// 11 -|RS           E|- 14
// 12 -|Vcc        R/W|- 13
//     +--------------+ 
//
// address 0 read	 = data bits 0-7
// address 0 write   = data bits 0-7
//
// address 1 write
// DB0 = divide sel 0     00 = /1	 10 = /64
// DB1 = divide sel 1	  01 = /16	 11 = master reset
// DB2 = word sel 0		 000 = 7e2	011 = 7o1		110 = 8e1
// DB3 = word sel 1		 001 = 7o2	100 = 8n2		111 = 8o1
// DB4 = word sel 2		 010 = 7e1	101 = 8n1
// DB5 = transmit ctrl 0 00 = -RTS low xmit irq disable
// DB6 = transmit ctrl 1 01 = -RTS low xmit irq enable
//                       10 = -RTS high xmit irq disable
//                       11 = -RTS low, xmit a break lvl, irq disable
// DB7 = receive irq enable
//
// address 1 read
// DB0 = RX buff full
// DB1 = TX buff empty
// DB2 = -DCD			(1 when DCD not present)
// DB3 = -CTS			(1 when transmit is inhibited)
// DB4 = Framing error	(1 when error, lack of start/stop bits or break condition)
// DB5 = RX overrun
// DB6 = Parity Error   (1 when # of 1 bits != selected parity)
// DB7 = IRQ            (1 when IRQ present)  cleared by read of rx or write to tx

// rxIRQ is set when Receive Data Register is full, overrun, or there is a low to high 
// transition on the DCD signal line.
//
// txIRQ is set when the Transmit Data Register is empty,

#define acia_IRQ	 '\x80'
#define acia_PE		 '\x40'
#define acia_RXO	 '\x20'
#define acia_FE		 '\x10'
#define acia_CTS	 '\x08'
#define acia_DCD     '\x04'
#define acia_TXEmpty '\x02'
#define acia_RXFull  '\x01'

class Acia
{
public:
	Acia( int nUnit =0 );
	virtual ~Acia();


	enum eWordSel {SevenE2 = 0,	SevenO2,SevenE1,SevenO1,EightN2,EightN1,EightE1,EightO1	};
	enum eDivideSel { DivideBy1, DivideBy16, DivideBy64, ResetACIA };
	enum eXmitCtrl { RTSNOIRQ, RTSIRQ, NORTSNOIRQ, RTSBREAK };
	enum eStatus { rxFull = 1,
				   txFull = 2,
				   inDCD = 4,
				   inCTS = 8,
				   FrameError = 16,
				   rxOverrun = 32,
				   rxParityError = 64,
				   rxIRQ = 128 };
	enum eBaud { b110=91,b300=33,b600=17,b1200=8,b2400=4,b4800=2,b9600=1,bNoLimit=0 };

//implementation
public:

	//acia access based on memory address
	void Write( unsigned int address, unsigned char writevalue);
	void Read( unsigned int address, unsigned char *readvalue);

	//acia access to individual functions
	void control_write(BYTE data);
	BYTE control_read();
	void data_write(BYTE data);
	BYTE data_read();
	void clock_tick();	

	//support functions
	void SetFile(BOOL bInput = TRUE);			//prompt use for input or output file
	void CloseFile(BOOL bInput = TRUE);			//close file if open
	BOOL IsFileOpen(BOOL bInput = TRUE);		//determine open state

	void SetBaud(eBaud Speed);					//set i/o pacing
	void SetINIFile(LPCTSTR szFile);			//Set .ini file
	BOOL LoadConfig( LPCTSTR szFile = NULL );	//Read .ini file settings
	BOOL SaveConfig( LPCTSTR szFile = NULL );	//Save .ini file settings
	
	//Accessors
	eBaud GetBaud() { return m_eSpeed; };
	BOOL  GetAutoSave() { return m_bAutoSave; };
	void  SetAutoSave(BOOL bSet) { m_bAutoSave = bSet; };
	BOOL  GetAutoLoad() { return m_bAutoLoad; };
	void  SetAutoLoad(BOOL bSet) { m_bAutoLoad = bSet; };

	//configuration dlgproc
	BOOL static CALLBACK SerialDlg(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam);
protected:

	void Reset();

protected:
	int		m_nUnit;	//Unit number of this ACIA 
	
	HANDLE hInFile;
	HANDLE hOutFile;

	TCHAR m_szIniFile[_MAX_PATH];

	eBaud   m_eSpeed;
	
	//mirror control registers
	eDivideSel m_eDivideSel;
	eXmitCtrl  m_eXmitCtrl;
	eWordSel   m_eWordSel;

	//transmit/receive pacing
	DWORD  m_dwRTickCount;
	DWORD  m_dwRNextTick;
	DWORD  m_dwTTickCount;
	DWORD  m_dwTNextTick;


	BYTE   m_cOut;		//last character received
	BYTE   m_cStatus;	//control status read

	BOOL   m_txIRQ;			//transmit buffer empty IRQ
	BOOL   m_rxIRQ;			//receive buffer full IRQ

	BOOL   m_txIRQenable;	//receive irq enabled
	BOOL   m_rxIRQenable;	//transmit irq enabled

	BOOL	m_bAutoSave;
	BOOL	m_bAutoLoad;
	

};

#endif // !defined(AFX_ACIA_H__EE7A53A0_A454_11D5_B5C9_00E01890870B__INCLUDED_)
