#include "w_osi.h"
#include "disk_io.h"
#include "6502.h"
#include "fileio.h"
#include "monitor.h"
#include "resource.h"

/////////////////////////////////////////////////////////////////////////////
//Port A
#define INDEX_D  0x80
#define DRVSEL_L 0x40
#define WR_PROT  0x20
#define DRV_RDY  0x10
#define TRACK0   0x02
#define DRV0_RDY 0x01

//Port B
#define WRITE_D  0x01
#define DIR      0x04
#define STEP     0x08
#define DRVSEL_H 0x20
#define HEAD_L	 0x80

DiskCtrl* DiskCtrl::m_gDisk = NULL;

/* OSI OS65D disk image */
//
// 5.25" disk, 40 tracks, 8  sector/track,  256 bytes/sector, 11 bits/byte (8E1) =80K/disk
// OSI uses 8E1 to give a max unformatted capacity of 2272 bytes per track (see below)
// However other bit encodings  (8N1) could give up to 2500 bytes/track 
//
// The standard speed for mini-floppy drives (5") is 300 rpm. Thus one 
// rotation of the disk is 200 mS.
//
// Stated baud-rate is 125k or 125000 bits/sec and one serial byte is 11 bits 
// (1 start, 8 data, 1 parity, 1 stop).
//
// So the theoretical absolute maximum storage per track is -
//    (125000 x 0.2) / 11  =  2272 bytes  or 8.8 pages
//
// OS-65D loses a bit more because it doesn't write until 10 mS after the 
// index pulse, so -
//    (125000 x 0.19) / 11  =  2159 bytes  or 8.4 pages
// and this doesn't even allow for the length of the index pulse (a few mS)? 
// and the speed variation between drives.
//---------------------------------------------------------------------------

// 8" disk, 77 tracks, 12 sectors/track, 256 bytes/sector, 11 bits/byte (8E1) = 231K/disk
// OSI uses 8E1 to give a max unformatted capacity of 3772 bytes/track (see below)
// However other bit encodings (8N1) could give up to 3900 bytes/track
//
// The standard speed for 8" floppy dives is 360 rpm. Thus one rotation of 
// the disk is 166.6ms
//
// Stated baud rate is 250K or 250000 bits/sec and one serial byte is 11 bits
// (1 start, 8 data, 1 parity, 1 stop).
//
// So the theoretical absolute maximum storage per track is -
//    (250000 x 0.166 ) / 11 = 	3772 or 14.7 pages
//
// OS-65D loses a bit more because it doesn't write until 10 mS after the 
// index pulse, so -
//    (250000 x 0.156) / 11  =  3545 bytes  or 13.8 pages
// and this doesn't even allow for the length of the index pulse (a few mS)? 
// and the speed variation between drives.
//---------------------------------------------------------------------------
// Track 0 format
// 10 millisecond delay after index hole
// 0,1 load address of the track in hi,lo form
// 2, page count of how much data is written on track 0.
// 3+ data
// 
// Track 1-n Format
// 10 ms delay after index hole
// 0,1 -a 2 byte start code $43, $57
// 2,  -BCD track number
// 3,  -track type code (always $58)
// (sector data follows)
//
// There can be any mixture of various length sectors. The total page count can not
// exceed 8 pages (8*256) if more than one sector is on a track
// Each sector is written in the following format:
// -previous sector length (4 if none before) times 800 microseconds of delay
// -sector start code $76
// -sector number in binary
// -sector length (#pages) in binary
// -sector data
// -(end of sector mark? $47, $53? MDS)
//
// Directory Format
// 2 sectors (1 & 2) on track 12 hold the directory information.
// Each entry requires 8 bytes. There are a total of 64 entries. The entries are
// formatted as follows:
// 0-5 ASCII 6 character filename
// 6 BCD first track of file
// 7 BCD Last track of file
//
// 
//
//
//
const MAXTRK5 =  39;		// maximum track number
const TRKSIZ5 = 0x0900;	// image track size (less for a real disk)

const MAXTRK8 = 76;	    // maximum tack number
const TRKSIZ8 = 0x0F00;  // image track size (less for real disk)


const TCHAR szDiskSection[]	= _T("DISK");
const TCHAR szPathEntry[]   = _T("LastPath");
const TCHAR szDiskCtrl[]    = _T("Disk Controller");
const TCHAR szDisk5Ext[]    = _T(".65d");
const TCHAR szDisk8Ext[]    = _T(".65u");


int disk_1s=0, disk_2s=0;


// OSI 505 board
// C000-C003 = PIA (decoded to C000-C00F) (C000/C040/C080/C0C0)
// C010-C011 = ACIA (Decode to C010-C01F) (C010/C050/C090/C0D0)
// C020-C03F = NULLREAD (C020/C060/C0A0/C0E0)

// PA0 <- !Drive 1 Ready		(0 = drive 1 ready)
// PA1 <- !TRACK 00				(0 = head on trk 0)
// PA2 <- !FAULT (nc)
// PA3 <- !SECTOR (nc)
// PA4 <- !Drive 2 Ready		(0 = drive 2 ready)
// PA5 <- !Write Protect		(0 = write protect)
// PA6 -> Drive Select			(0 = drv 2, 1 = drv 1)
// PA7 <- !INDEX
//
// PB0 -> !WRITE ENABLE			(0 = write)
// PB1 -> !ERASE ENABLE (nc)	(leave 1)
// PB2 -> !STEP IN				(0 = to trk 39)
// PB3 -> !RESET STEP 		    (step on 1->0 transition)
// PB4 -> !FAULT RESET (nc)		(leave 1)
// PB5 -> Drive Enable			(1 = drive ON) (Or Drive/Side sel. 4 drive system 1=drive 0&1,0=drive 2&3 )
// PB6 -> !LOW CURRENT (nc)		(leave 1)
// PB7 -> !HEAD LOAD			(0 = head on disk)
//
// CA1
// CA2
// CB1 <- 400 mSec clk for RTC operations (2.5x/sec) 5.25" disk
// CB2 
//

// Disk routines set port A to all input except PA6, port B to all output

DiskCtrl::DiskCtrl(BOOL b8Drive /*=FALSE*/, int nNumDrives /*=2*/) : 
	m_b8Inch(b8Drive), m_nNumDrives(nNumDrives)
{
	int i;

	ZeroMemory( &m_Drive, sizeof(m_Drive[0]) * MAX_DRIVE);
	for (i = 0; i < MAX_DRIVE; i++)
	{
		m_Drive[i].nIndex = 20;
		m_Drive[i].nTrack = 1;
	}

	m_nSelDrive	= 0;
	m_nDriveSel = 0;
	m_AciaData = -1;
	m_gDisk = NULL;

	CreateBuffer();

	m_nEvent = SetTimer(NULL, 0, 400, &DiskTimerProc );
	if (m_nEvent)
		m_gDisk = this;
}

DiskCtrl::~DiskCtrl()
{
	if (m_nEvent)
	{
		KillTimer( NULL, m_nEvent);
		m_nEvent = 0;
	}
	FreeBuffer();


}

//
// Callback to handle hardware 400ms timer on CB1, used for RTC
//
VOID CALLBACK DiskCtrl::DiskTimerProc(
  HWND ,     // handle of window for timer messages
  UINT ,     // WM_TIMER message
  UINT ,  // timer identifier
  DWORD    // current system time
)
{
	//TRACE("tick ");
	DiskCtrl* pCtrl = DiskCtrl::GetDiskPtr();
	if (pCtrl)
	{
		pCtrl->SetCB1(1);
	}
}
 

void DiskCtrl::CreateBuffer()
{
	
	long nTrack = 1 + (m_b8Inch ? MAXTRK8 : MAXTRK5);
	long nSize =  m_b8Inch ? TRKSIZ8 : TRKSIZ5;
	int i = 0;
	//allocate buffer for disk image(s)

	for ( i = 0; i < m_nNumDrives; i++ )
	{

		m_Drive[i].pDiskBuf = (BYTE*) HeapAlloc( GetProcessHeap(), HEAP_ZERO_MEMORY , nSize * nTrack );
		m_Drive[i].pDataPtr	= m_Drive[i].pDiskBuf;

		if (!m_Drive[i].pDiskBuf)
		{
			MessageBox(NULL, "Error allocationg memory for disk image", "Error", MB_OK);
		} 

	}	

}

void DiskCtrl::FreeBuffer()
{

	int i;
	
	for ( i = 0; i < MAX_DRIVE; i++ )
	{

		if (m_Drive[i].pDiskBuf)
			HeapFree(GetProcessHeap(), 0, m_Drive[i].pDiskBuf);
		m_Drive[i].pDiskBuf = NULL;
		m_Drive[i].pDataPtr = NULL;


	}	

}

//Set .ini file for other operations
void DiskCtrl::SetINIFile(LPCTSTR szFile)
{
	*m_szIniFile = _T('\0');
	if (_tcslen(szFile) < _MAX_PATH)
		_tcscpy( (LPTSTR) m_szIniFile, szFile);

}

BOOL DiskCtrl::SetDriveConfig( BOOL b8Inch, int nNumDrives)
{
	int i;

	long nTrack = 1 + (m_b8Inch ? MAXTRK8 : MAXTRK5);
	long nSize =  m_b8Inch ? TRKSIZ8 : TRKSIZ5;
	
	if (nNumDrives < 0 || nNumDrives > MAX_DRIVE)
		return FALSE;
	
	if (b8Inch != m_b8Inch)
	{
		
		//close all disks, reopen new sizes
		CloseFile(0);
		FreeBuffer();
		m_nNumDrives = nNumDrives;
		m_b8Inch = b8Inch;

		CreateBuffer();

		return TRUE;
	}

	if (nNumDrives < m_nNumDrives)
	{
		for (i = m_nNumDrives; i > nNumDrives; i--)
		{
			CloseFile(i);
			if (m_Drive[i].pDiskBuf)
				HeapFree(GetProcessHeap(), 0, m_Drive[i].pDiskBuf);
			m_Drive[i].pDiskBuf = NULL;
			m_Drive[i].pDataPtr = NULL;
		}
	}
	else if (nNumDrives > m_nNumDrives)
	{
		for (i = m_nNumDrives +1; i <= nNumDrives; i++)
		{
			
			if (!m_Drive[i-1].pDiskBuf)
			{

				m_Drive[i-1].pDiskBuf = (BYTE*) HeapAlloc( GetProcessHeap(), HEAP_ZERO_MEMORY , nSize * nTrack );
				m_Drive[i-1].pDataPtr	= m_Drive[i-1].pDiskBuf;

				if (!m_Drive[i-1].pDiskBuf)
				{
					MessageBox(ghWnd_Main, "Error allocationg memory for disk image", "Error", MB_OK);
				}
			
			} 
		}
	}

	return TRUE;

}

#define TITLELEN 80
#define MAXFILTER 255
void DiskCtrl::OpenFile(int nDrive, TCHAR* szFileArg)
{

	OPENFILENAME ofn;
	TCHAR file[_MAX_PATH];
	TCHAR szTitle[TITLELEN];
	TCHAR szFilter[MAXFILTER];
	TCHAR Path[_MAX_PATH];
	BOOL bOK = TRUE;
	HANDLE hInFile = INVALID_HANDLE_VALUE;
	DWORD dwAttr = 0;
	
	long nTracks = 1 + (m_b8Inch ? MAXTRK8 : MAXTRK5);
	long nTrkSize =  m_b8Inch ? TRKSIZ8 : TRKSIZ5;
	int nVal = 1 << (nDrive-1);
	*file = _T('\0');

	if (nDrive > MAX_DRIVE || nDrive < 1)
		return;

	m_nSelDrive &= ~nVal;
	
	if (GetPrivateProfileString(szDiskSection, szPathEntry, _T(""), 
								Path, _MAX_PATH, m_szIniFile) < 2)
	{
		_tcscpy(Path, sztFileHome);
	}
	else
	{
		int nLen = _tcslen(Path);
		if (Path[nLen-1] != _T('\\'))
		{
			//determine if entry is just missing trailing '\' or is filename
			dwAttr = GetFileAttributes(Path);
			if (dwAttr != (DWORD) -1 && (dwAttr & FILE_ATTRIBUTE_DIRECTORY))
			{
				//it's a dir, add trailing '\'
				_tcscat(Path, _T("\\"));
			}
			else
			{
				//aint a dir, maybe gone or just a file, chop at last '\'
				TCHAR *p = _tcsrchr(Path, _T('\\'));
				if (p)
					*(p+1) = _T('\0');
			}
		}

	}

	ZeroMemory(&ofn, sizeof(OPENFILENAME));
	
	if (!szFileArg || !*szFileArg)
	{

		if (LoadString(ghInstance, IDS_DISKFILTER, szFilter, MAXFILTER))
		{
			for (int i=0; i < MAXFILTER; i++)
			{
				//filter needs null separating paramters, replace | in resource with Null
				if (szFilter[i] == _T('|'))
					szFilter[i] = _T('\0');
			}
		} 


		*file = _T('\0');
		*szTitle = _T('\0');


		LoadString(ghInstance, IDS_DISK_TITLE, szTitle, TITLELEN);
		

		ofn.lStructSize = sizeof(OPENFILENAME);
		ofn.hwndOwner = ghWnd_Main;
		ofn.hInstance = ghInstance;
		ofn.Flags = OFN_EXPLORER |  OFN_FILEMUSTEXIST | OFN_ENABLETEMPLATE | OFN_ENABLEHOOK | OFN_ENABLESIZING;
		ofn.lpstrInitialDir	= Path;
		ofn.lpstrFile = file;
		ofn.nMaxFile = _MAX_PATH;
		ofn.lpstrTitle = szTitle;
		ofn.lpstrFilter = szFilter;
		ofn.lpTemplateName = MAKEINTRESOURCE(IDD_FILEDLG);
		ofn.lpfnHook = FileHookProc;

		bOK = GetOpenFileName( &ofn );
	}
	else
	{
		
		if ( _tcschr(szFileArg,_T('\\')) == NULL)
		{
			_tcscpy(file, Path);
		}
		_tcscat(file, szFileArg);
	}

	*m_Drive[nDrive-1].szFile = _T('\0');
	
	if (bOK)
	{
		
		hInFile = CreateFile(file, GENERIC_READ, FILE_SHARE_READ,
				NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, 
				NULL );

		if (hInFile == INVALID_HANDLE_VALUE)
		{
			ReportSystemError(GetLastError());
			return;
		}
		else
		{
			DWORD dwRead = 0, dwSize = nTrkSize * nTracks;	
			bOK = ReadFile( hInFile, m_Drive[nDrive-1].pDiskBuf, dwSize, &dwRead, NULL);

			if (!bOK || dwRead != dwSize)
			{
				if (bOK)
					MessageBox(ghWnd_Main, "Error reading file for disk image", "Error", MB_OK);
				else
					ReportSystemError(GetLastError());
				bOK = FALSE;
			}

			if (bOK)
			{
				
				TCHAR drive[_MAX_DRIVE];

				m_nSelDrive |= nVal;
				
				//Set read-only (Write protect) flag
				dwRead = GetFileAttributes(file);
				if ( dwRead != (DWORD) -1 && (dwRead & FILE_ATTRIBUTE_READONLY) != 0)
					m_Drive[nDrive-1].bWProtect = TRUE;
				else
					m_Drive[nDrive-1].bWProtect = FALSE;


				//save opened disk image filename
				_tcscpy(m_Drive[nDrive-1].szFile, file);

				_stprintf(Path,_T("%s%c"), _T("DRIVE_"), _T('A')+nDrive-1);

				//save filename for reloading
				WritePrivateProfileString(szDiskSection, Path, file, m_szIniFile);

				//save last load path, remove filename
				_tsplitpath(file, drive, Path, NULL, NULL);
				_tmakepath(file, drive, Path , NULL, NULL);
			}

			CloseHandle(hInFile);
			hInFile = INVALID_HANDLE_VALUE;
		}

	}
	if (bOK && m_szIniFile && *m_szIniFile)
	{
		//save open path
		WritePrivateProfileString(szDiskSection, szPathEntry, file, m_szIniFile);

	}

}

void DiskCtrl::CloseFile(int nDrive /*= 0*/)
{
	int i;

	if (nDrive > MAX_DRIVE || nDrive < 0)
		return;

	if (nDrive == 0)
	{
		//close all files

		for ( i = 0; i < MAX_DRIVE; i++)
		{
			if (m_Drive[i].bChanged)
			{
				//Prompt to save changed disk image
				SaveFile(i+1);
			}

			*m_Drive[i].szFile = _T('\0');

			m_nSelDrive &= ~(1 << i);

		}
	}
	else
	{
		//close specific file
		if (m_Drive[nDrive-1].bChanged)
		{
			//Prompt to save changed disk image
			SaveFile(nDrive);
			m_Drive[nDrive-1].bChanged = FALSE;
			m_nSelDrive &= ~(1 << (nDrive-1));
		}

		*m_Drive[nDrive-1].szFile = _T('\0');

	}


}

void DiskCtrl::SaveFile(int nDrive, BOOL bPrompt /*=TRUE*/)
{
	
	TCHAR szMessage[500];

	HANDLE hFile = INVALID_HANDLE_VALUE;
	BOOL bOK = TRUE;
	long nTracks = 1 + (m_b8Inch ? MAXTRK8 : MAXTRK5);
	long nTrkSize =  m_b8Inch ? TRKSIZ8 : TRKSIZ5;


	if (nDrive > MAX_DRIVE || nDrive <= 0)
		return;

	_stprintf(szMessage, _T("Save modified disk image in Drive %c?"), _T('@')+nDrive); 

	
	if ( !bPrompt || MessageBox( ghWnd_Main, szMessage, szDiskCtrl, MB_YESNO) == IDYES)
	{
		DWORD dwWritten;
		
		hFile = CreateFile(m_Drive[nDrive-1].szFile, GENERIC_WRITE | GENERIC_READ, FILE_SHARE_READ,
			NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );

		if (hFile == INVALID_HANDLE_VALUE)
		{
			ReportSystemError(GetLastError());
			return;
		}
		bOK = FALSE;
		while (!bOK)
		{
			DWORD dwWrite = nTracks * nTrkSize;
			
			bOK = WriteFile( hFile, m_Drive[nDrive-1].pDiskBuf, nTracks * nTrkSize, &dwWritten, NULL);
		
			if (!bOK || dwWrite != dwWritten)
			{
				
				LPTSTR   message; 
 
				FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER | 
							   FORMAT_MESSAGE_FROM_SYSTEM, 
							   NULL, 
							   GetLastError(), 
							   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), 
							   (LPTSTR) &message, 
							   0, 
							   NULL); 
 
				_stprintf(szMessage, _T("Error writing file %s.\r\n%s"), m_Drive[nDrive-1].szFile, message);

				LocalFree (message);				
				
				if (MessageBox(ghWnd_Main, szMessage, szDiskCtrl, MB_RETRYCANCEL) == IDCANCEL)
					bOK = TRUE;
					
			}
		}

		CloseHandle(hFile);
		hFile = INVALID_HANDLE_VALUE;

	}

}

//virtual function called from base class to handle hardware simulation
void DiskCtrl::WritePortA(BYTE writevalue)
{
	m_nDriveSel = 0x03;
	
	// Check actual selected drive from PB5,PA6
	if ((port_b & DRVSEL_H) == DRVSEL_H) 
		m_nDriveSel ^= 2;	// PB5   Side Select
	
	if ((writevalue & DRVSEL_L) == DRVSEL_L) 
		m_nDriveSel ^= 1; // PA6   Boot Drw=1
	
	m_nDriveSel++;
	
	//	wsprintf(string,TEXT("PORT A  DRV=%02X  writevalue=%02X"),drv_sel,writevalue);
	//	MessageBox(ghWnd_Main, string,TEXT("C000 Write"), MB_ICONSTOP);

}
BYTE DiskCtrl::ReadPortA()
{
	
	BYTE Aport = 0xFF;			   // PA2=PA3=1

	Aport ^= DRV0_RDY;				// PA0=0
	Aport ^= DRV_RDY;				// Set PA4=0 Ready

	if ((port_a & DRVSEL_L) != DRVSEL_L)
		Aport ^= DRVSEL_L;  // Set PA6 info
	
	if (m_nDriveSel)
	{
		int nDrive = m_nDriveSel-1;
		int nSel = 1 << nDrive;

		if ( (m_nSelDrive & nSel) != 0)
		{
			//disk_1s=1;
			Aport &= ~WR_PROT;
			
			if (!m_Drive[nDrive].bWProtect)
				Aport |= WR_PROT;	//not protected
			
			if (m_Drive[nDrive].nTrack == 0) 
				Aport ^= TRACK0;	// TR0 to 0
			
			m_Drive[nDrive].nIndex--;				// Simulate Index signal
			
			if (m_Drive[nDrive].nIndex < 11)			// Index stays 1
			{
				if (m_Drive[nDrive].nIndex > 0)
				{
					Aport ^= INDEX_D;  			// Set Index to 0
					m_AciaCtrl=0x0E;			// During index=low no Data
					m_Drive[nDrive].nSector=0;
					//disk_1s=0;
				}
				else
				{
					m_AciaCtrl=0x0F;			// After end of index=0 Data available
					m_Drive[nDrive].nIndex = 100;	// Index block after Track up/down
				}
			}

			// track size for OS65D disk image
			// set pointer on track offset TR0...
			
			//This is done when trk stepped MDS
			//if (m_Drive[i].pDiskBuf)
			//	m_Drive[i].pDataPtr= m_Drive[i].pDiskBuf + (m_Drive[i].nTrack * (m_b8Inch ? TRKSIZ8 : TRKSIZ5));
			//else
			//	m_Drive[i].pDataPtr = NULL;
		
		}

	}  // ENDIF Step only on drive sel
	
	return Aport;
}

void DiskCtrl::WritePortB(BYTE writevalue)
{
	// check changes on portbdat
	m_nDriveSel = 0x03;

	// Check actual selected drive from PB5,PA6
	if ( (writevalue & DRVSEL_H) == DRVSEL_H) 
		m_nDriveSel^=2;	// PB5   Side Select

	if ( (port_a & DRVSEL_L) == DRVSEL_L) 
		m_nDriveSel^=1;		// PA6
	
	m_nDriveSel++;

	//		wsprintf(string,TEXT("ADR=%04X  DRV=%02X Index=%02X  PAL=%02X  PBH=%02X  <>  TRA=%02X TRB=%02X"),R.PC.W,drv_sel,drv_index,(portadat&DRVSEL_L),(writevalue&DRVSEL_H),track_a,track_b);
	//		MessageBox(ghWnd_Main, string,TEXT("C002 Write"), MB_ICONSTOP);
	//		CIA_Clear();

	if ( (writevalue & HEAD_L) == HEAD_L)	// PB7 Sel is always active on Head_L=0
	{
		//Head not loaded or going to be unloaded
		
		// Check if Step is high
		if ( (writevalue & STEP) == STEP)	// PB3 During Step low Drive is always selected
		{
			//Step not stepping
			
			// Check Head Load pulse going to high
			if (((port_b & HEAD_L) == 0) )
			{
				//head unloaded now
				m_nDriveSel = 0;	// FF clk -> Sel no longer active
			}

		} // Endif Step=0
	
	} // Endif Head_L=0

	if (m_nDriveSel > 0)
	{
		int nDrive = m_nDriveSel-1;
		int nSel = 1 << nDrive;

		if ( (m_nSelDrive & nSel) != 0 )
		{
			if ( (port_b & STEP) && !(writevalue & STEP) )	// Step pulse low edge
			{
				//step pulse going from hi to lo, do step
				
				if ( writevalue & DIR)	//H direction = towards trk 0
					m_Drive[nDrive].nTrack--;
				else
					m_Drive[nDrive].nTrack++;
				
				m_Drive[nDrive].nIndex = 20;              			// initial index pulse
				m_Drive[nDrive].nSector = 0;

				port_a |= INDEX_D;				// Set Index to 1
				//	wsprintf(string,TEXT("drv_sel=%02X track_a=%02X"),drv_sel,track_a);
				//	MessageBox(ghWnd_Main, string,TEXT("C002 Write"), MB_ICONSTOP);
			
				if (m_Drive[nDrive].nTrack > ( m_b8Inch ? MAXTRK8 : MAXTRK5) )
					m_Drive[nDrive].nTrack = ( m_b8Inch ? MAXTRK8 : MAXTRK5 );
				
				if (m_Drive[nDrive].nTrack < 0) 
					m_Drive[nDrive].nTrack = 0;

				//set data ptr
				m_Drive[nDrive].pDataPtr = m_Drive[nDrive].pDiskBuf + 
								(m_Drive[nDrive].nTrack * (m_b8Inch ? TRKSIZ8 : TRKSIZ5));
			
			} //If stepped

		} //if drive image selected
	
	} //if drive selected
}

BYTE DiskCtrl::ReadPortB()
{
	return port_b;
}

// OSI 505 board
// C000-C003 = PIA (decoded to C000-C00F) (C000/C040/C080/C0C0)
// C010-C011 = ACIA (Decode to C010-C01F) (C010/C050/C090/C0D0)
// C020-C03F = NULLREAD (C020/C060/C0A0/C0E0)

void DiskCtrl::Write( unsigned int address, unsigned char writevalue)
{
	address &= 0x3F;

	if (address < 0x10)
	{
		//let base class handle it
		PIA6821::Write( address, writevalue);
	}
	else if (address < 0x20)
	{
		if ( (address & 0x01) == 1)
		{
			//ACIA Data Write
			TRACE("DiskACIA DATA Write 0x%x\n", writevalue);
		}
		else
		{
			TRACE("DiskACIA CTRL Write 0x%x\n", writevalue);
			//ACIA Ctrl Write
			if (writevalue == 0x03)
			{
				m_AciaData = -1;
				m_AciaCtrl = 0x0E;
			}
			else
			{
				m_AciaData = 0xFF;
				m_AciaCtrl = 0x0E;
			}
		}
	}
	//else NULLWRITE

}

// $C000 = DDRA/PORTA
// $C001 = PORT A CTRL
// $C002 = DDRB/PORTB
// $C003 = PORT B CTRL
// $C010 = ACIA CTRL
// $C011 = ACIA DATA
void DiskCtrl::Read( unsigned int address, unsigned char *readvalue)
{

	int nDrive = m_nDriveSel-1;

	if ( (address & 0x3F) < 0x10)
	{
		//let base class handle it
		PIA6821::Read( address, readvalue);
	}
	else if ( (address & 0x3F) < 0x20)
	{
		if ( (address & 0x01) && m_nDriveSel)
		{
			//ACIA Data Read
			if( m_AciaCtrl == 0x0F )
			{
				int nSel = 1 << nDrive;
				

				if (m_nSelDrive & nSel)	//Is file image loaded?
				{
					
					if (!m_Drive[nDrive].pDataPtr)
					{
						MessageBox(NULL, "Error in disk controller", "Error", MB_OK);	
					}
					else
					{
						//read next byte
						m_AciaData =  *(m_Drive[nDrive].pDataPtr + m_Drive[nDrive].nSector);
					}
					m_Drive[nDrive].nSector++;
					
					//if (m_Drive[nDrive].nSector >= ( m_b8Inch ? TRKSIZ8 : TRKSIZ5 )  )
					//{
					//	m_Drive[nDrive].nSector = 0;
					//	m_AciaCtrl = 0x0E; //Set No More data?
					//}
					
					
					
				//	wsprintf(string,TEXT("drv_sel=%02X  Pos=%04X  dr_in=%02X Aciac=%02X"),drv_sel,sector_a-1,dr_in&0xff,*acia2cptr);
				//	MessageBox(ghWnd_Main, string,TEXT("C011 File A Read"), MB_ICONSTOP);
				}

			}
			else
			{ 
				m_AciaData=0xFF; 					// Disk RDA Data high
			}
			
			m_Drive[nDrive].nIndex = 100;

	
			if (m_Drive[nDrive].nSector >= ( m_b8Inch ? TRKSIZ8 : TRKSIZ5 ) )
			{
				m_Drive[nDrive].nSector = 0;
				m_AciaCtrl = 0x0E;					//Set No More data available?
				m_Drive[nDrive].nIndex = 10;	// force index pulse on next poll of $C000
			}
			
			*readvalue = (BYTE) m_AciaData;
		}
		else 
		{
			//ACIA Ctrl Read
			if (m_nDriveSel)
				m_AciaCtrl=0x0F;
			else
				m_AciaCtrl=0x0E;
			*readvalue=m_AciaCtrl;
		}
	}
	else
	{
		NULLREAD(); 
	}

}
void DiskCtrl::Reset()
{
	CloseFile(0);

	PIA6821::Reset();
		
}

UINT CALLBACK DiskCtrl::FileHookProc(HWND hdlg, UINT uiMsg, WPARAM wParam, LPARAM lParam )
{

	HWND hList = NULL;
	TCHAR filename[_MAX_PATH];
	static HFONT hfont=NULL;
	WORD msg_type;

    hList =GetDlgItem(hdlg,IDC_LB_PREVIEW);

    switch (uiMsg) 
	{
		case WM_INITDIALOG:
			//SetWindowText(GetDlgItem(GetParent(hwnd),IDOK),"&Attach");
			

			//create a fixed width font for the listbox
			hfont = CreateFont(-12,-7,0,0,400,0,0,0,0,0,0, DRAFT_QUALITY,
							   FIXED_PITCH|FF_MODERN,NULL);
			if (hfont) 
			{
				SendDlgItemMessage(hdlg,IDC_EB_PREVIEW, WM_SETFONT,
					(WPARAM)hfont,MAKELPARAM(TRUE,0));
			}

			SetWindowText(GetDlgItem(hdlg, IDC_EB_PREVIEW),"File name   Track range\r\n-----------------------\r\n OS65D3      0  - 13\r\n *BEXEC      14 - 16");

			break;
		case WM_NOTIFY:
			if ( ((OFNOTIFY*)lParam)->hdr.code == CDN_SELCHANGE) 
			{
				//new file selected from file list
				SendMessage(hList,LB_RESETCONTENT,0,0);
				if (SendMessage(((OFNOTIFY*)lParam)->hdr.hwndFrom,
					CDM_GETFILEPATH,_MAX_PATH ,(LPARAM)filename)>=0) 
				{
					 if (!(GetFileAttributes(filename)&FILE_ATTRIBUTE_DIRECTORY)) 
					 {
						 
						 /*
						 if (read_content_func!=NULL) 
						 {
							 contents=read_content_func(filename);
							 create_content_list(contents,preview);
						 }
						 */
					 }
				}
			} 
			else if (((OFNOTIFY*)lParam)->hdr.code == CDN_FOLDERCHANGE) 
			{
				SetWindowText(GetDlgItem(hdlg, IDC_EB_PREVIEW),"");
				SetWindowText(GetDlgItem(GetParent(hdlg),0x0480),"");
				SetWindowText(GetDlgItem(hdlg, IDC_EB_PREVIEW),"File name   Track range\r\n-----------------------\r\n OS65D3      0  - 13\r\n *BEXEC      14 - 16");

			}
			break;
		case WM_COMMAND:
			msg_type = LOWORD(wParam);
			switch (msg_type) 
			{
				case IDC_PB_CREATENEW:
					if (SendMessage(GetParent(hdlg),
						CDM_GETSPEC, _MAX_PATH, (LPARAM)filename) <= 1) 
					{
						MessageBox(NULL,"","Please enter a filename.", MB_OK);
						return (UINT)-1;
					}

					if (SendMessage(GetParent(hdlg),
						CDM_GETFILEPATH,_MAX_PATH,(LPARAM)filename)>=0)
					{
						/*
						char disk_name[32];
						char disk_id[3];
						char *format_name;

						counter = SendMessage(GetDlgItem(hwnd,IDC_BLANK_IMAGE_TYPE),
													 CB_GETCURSEL,0,0);
						if (append_extension) 
						{
							strcat(filename,".");
							strcat(filename,image_type_name[counter]);
						}
						if (util_file_exists_p(filename)) 
						{
							int ret;
							ret = ui_messagebox("Overwrite existing image?",
								"VICE question", MB_YESNO | MB_ICONQUESTION);
							if (ret != IDYES)
								return -1;
						}
						GetDlgItemText(hwnd,IDC_BLANK_IMAGE_NAME, disk_name, 17);
						GetDlgItemText(hwnd,IDC_BLANK_IMAGE_ID, disk_id, 3);
						format_name = xmsprintf("%s,%s", disk_name, disk_id);
						if (vdrive_internal_create_format_disk_image(filename,
							format_name, image_type[counter]) < 0) 
						{
							ui_error("Cannot create image");
							free(format_name);
							return -1;
						}
						free(format_name);
						//  Select filter:
						//If we have a standard extension, select the disk filters,
						//but leave at 'All files' if it was already there, otherwise
						//select 'All files'
						//
						
						index = SendMessage(GetDlgItem(GetParent(hwnd), 0x470),
										CB_GETCOUNT, 0, 0);
						if (is_it_standard_extension) 
						{
							if (index-1!=SendMessage(GetDlgItem(GetParent(hwnd),
													0x470), CB_GETCURSEL, 0, 0)) 
							{
								SendMessage(GetDlgItem(GetParent(hwnd), 0x470),
										CB_SETCURSEL, 0, 0);
							}
						} 
						else 
						{
							SendMessage(GetDlgItem(GetParent(hwnd), 0x470),
									CB_SETCURSEL, index - 1, 0);
						}
						//  Notify main window about filter change 
						SendMessage(GetParent(hwnd), WM_COMMAND,
									MAKELONG(0x470, CBN_SELENDOK),
									(LPARAM)GetDlgItem(GetParent(hwnd), 0x470));

						//  Now find filename in ListView & select it
						SendMessage(GetParent(hwnd), CDM_GETSPEC, 256,
									(LPARAM)filename);
						if (append_extension) 
						{
							strcat(filename, ".");
							strcat(filename, image_type_name[counter]);
						}
						find.flags = LVFI_STRING;
						find.psz = filename;
						index = SendMessage(GetDlgItem(GetDlgItem(GetParent(hwnd),
								0x461), 1), LVM_FINDITEM, -1, (LPARAM)&find);
						item.stateMask = LVIS_SELECTED | LVIS_FOCUSED;
						item.state = LVIS_SELECTED | LVIS_FOCUSED;
						SendMessage(GetDlgItem(GetDlgItem(GetParent(hwnd), 0x461), 1),
									LVM_SETITEMSTATE, index, (LPARAM)&item);
					*/
					}
				break;
			}
			switch (HIWORD(wParam)) 
			{
				case LBN_DBLCLK:
					/*
					if (autostart_result != NULL) 
					{
						index = SendMessage((HWND)lparam, LB_GETCURSEL, 0, 0);
						if (SendMessage(GetParent(hwnd),
							CDM_GETFILEPATH, 256, (LPARAM)filename) >= 0) 
						{
							*autostart_result = get_filename_from_content(filename,index);
							SendMessage(GetParent(hwnd), WM_COMMAND,
										MAKELONG(IDOK,BN_CLICKED),
										(LPARAM)GetDlgItem(GetParent(hwnd), IDOK));
						}
					}
					*/
					break;
			}
			break;
		case WM_DESTROY:
			if (hfont != NULL) 
			{
				DeleteObject(hfont);
				hfont = NULL;
			}
			break;
    }
    return 0;



} 

 

